@@ -72,10 +72,10 @@ The configuration must be in JSON format. I.e., something like
   }
 
 The keys and structure of this file depend on the type of the service
-you are setting up. "CORS" and "debug" are the only ones that are
-recognized by this module. "CORS" is either a string denoting the
-allowed origin or a hash of "Allow-Origin", "Allow-Methods",
-"Allow-Headers", and "Max-Age".
+you are setting up. "CORS" is the only one that is recognized by this
+module. "CORS" is either a string denoting the allowed origin or a
+hash of "Allow-Origin", "Allow-Methods", "Allow-Headers", and
+"Max-Age".
 
 =head2 EXPORT
 
@@ -89,12 +89,16 @@ package Geo::OGC::Service;
 
 use 5.022000;
 use Modern::Perl;
+use Encode qw(decode encode);
 use Plack::Request;
 use JSON;
 use XML::LibXML;
 use Clone 'clone';
+use XML::LibXML::PrettyPrint;
 
-our $VERSION = '0.02';
+binmode STDERR, ":utf8"; 
+
+our $VERSION = '0.03';
 
 =pod
 
@@ -106,7 +110,8 @@ parameters are
 
   configuration, services
 
-configuration is a path to a file.
+configuration is a path to a file or a reference to an anonymous hash
+containing the configuration.
 
 services is a reference to a hash of service names associated with
 names of classes, which will process those service requests.
@@ -156,25 +161,28 @@ while interpreting the request, or while processing the request.
 
 sub respond {
     my ($responder, $env, $config, $services) = @_;
-    if (open(my $fh, '<', $config)) {
-        my @json = <$fh>;
-        close $fh;
-        eval {
-            $config = decode_json "@json";
-        };
-        unless ($@) {
-            $config->{CORS} = $ENV{'REMOTE_ADDR'} unless $config->{CORS};
-            $config->{debug} = 0 unless defined $config->{debug};
+    if (not ref $config) {
+        if (open(my $fh, '<', $config)) {
+            my @json = <$fh>;
+            close $fh;
+            eval {
+                $config = decode_json "@json";
+            };
+            unless ($@) {
+                $config->{CORS} = $ENV{'REMOTE_ADDR'} unless $config->{CORS};
+                $config->{debug} = 0 unless defined $config->{debug};
+            } else {
+                print STDERR "$@";
+                undef $config;
+            }
         } else {
-            print STDERR "$@";
+            print STDERR "Can't open file '$config': $!\n";
             undef $config;
         }
-    } else {
-        print STDERR "Can't open file '$config': $!\n";
-        undef $config;
     }
     unless ($config) {
-        error($responder, 'Configuration error.');
+        error($responder, { exceptionCode => 'ResourceNotFound',
+                            ExceptionText => 'Configuration error.' } );
     } elsif ($env->{REQUEST_METHOD} eq 'OPTIONS') {
         my %cors = ( 
             'Content-Type' => 'text/plain',
@@ -196,21 +204,22 @@ sub respond {
         }
         $responder->([200, [%cors]]);
     } else {
-        my $request = Plack::Request->new($env);
         my $service;
         eval {
-            $service = service($responder, $request, $config, $services);
+            $service = service($responder, $env, $config, $services);
         };
         if ($@) {
             print STDERR "$@";
-            error($responder, "Error in interpreting the request.");
+            error($responder, { exceptionCode => 'ResourceNotFound',
+                                ExceptionText => "Internal error while interpreting the request." } );
         } elsif ($service) {
             eval {
                 $service->process_request($responder);
             };
             if ($@) {
                 print STDERR "$@";
-                error($responder, "Error in processing the request.");
+                error($responder, { exceptionCode => 'ResourceNotFound',
+                                    ExceptionText => "Internal error while processing the request." } );
             }
         }
     }
@@ -227,6 +236,7 @@ is merged into the object.
 The returned service object contains
 
   config => a clone of the configuration for this service
+  env => many values from the PSGI environment
 
 and may contain
 
@@ -234,33 +244,45 @@ and may contain
   filter => XML::LibXML DOM document element of the filter
   parameters => hash of rquest parameters obtained from Plack::Request
 
+Note: all keys in request parameters are converted to lower case in
+parameters.
+
 This subroutine may fail due to a request for an unknown service.
 
 =cut
 
 sub service {
-    my ($responder, $request, $config, $services) = @_;
+    my ($responder, $env, $config, $services) = @_;
 
+    my $request = Plack::Request->new($env);
     my $parameters = $request->parameters;
     
     my %names;
     for my $key (sort keys %$parameters) {
         $names{lc($key)} = $key;
-        print STDERR "request kvp: $key => $parameters->{$key}\n" if $config->{debug} > 2;
     }
 
     my $self = { config => clone($config) };
-    
+    for my $key (qw/SCRIPT_NAME PATH_INFO SERVER_NAME SERVER_PORT SERVER_PROTOCOL CONTENT_LENGTH CONTENT_TYPE
+                    psgi.version psgi.url_scheme psgi.multithread psgi.multiprocess psgi.run_once psgi.nonblocking psgi.streaming/) {
+        $self->{env}{$key} = $env->{$key};
+    };
+    for my $key (keys %$env) {
+        $self->{env}{$key} = $env->{$key} if $key =~ /^HTTP_/ || $key =~ /^REQUEST_/;
+    };
+
     my $post = $names{postdata} // $names{'xforms:model'};
+    $post = $post ? $parameters->{$post} : encode($request->content_encoding // 'UTF-8', $request->content);
 
     if ($post) {
         my $parser = XML::LibXML->new(no_blanks => 1);
         my $dom;
         eval {
-            $dom = $parser->load_xml(string => $parameters->{$post});
+            $dom = $parser->load_xml(string => $post);
         };
         if ($@) {
-            error($responder, "Error in posted XML:\n$@");
+            error($responder, { exceptionCode => 'ResourceNotFound',
+                                ExceptionText => "Error in posted XML:\n$@" } );
             return;
         }
         $self->{posted} = $dom->documentElement();
@@ -276,7 +298,8 @@ sub service {
                     $dom = $parser->load_xml(string => $filter);
                 };
                 if ($@) {
-                    error($responder, "Error in XML filter:\n$@");
+                    error($responder, { exceptionCode => 'ResourceNotFound',
+                                        ExceptionText => "Error in XML filter:\n$@" } );
                     return;
                 }
                 $self->{filter} = $dom->documentElement();
@@ -286,20 +309,38 @@ sub service {
         }
     }
 
-    my $service = $self->{parameters}{service} // '';
+    # service may also be an attribute in the top element of posted XML
+    my $service_from_posted = sub {
+        my $node = shift;
+        return undef unless $node;
+        return $node->getAttribute('service');
+    };
+
+    my $service = $self->{parameters}{service} // $service_from_posted->($self->{posted}) // ''; 
+
     if (exists $services->{$service}) {
         return bless $self, $services->{$service};
     }
 
-    error($responder, "Unknown service requested: '$service'.");
+    error($responder, { exceptionCode => 'InvalidParameterValue',
+                        locator => 'service',
+                        ExceptionText => "'$service' is not a known service to this server" } );
+    return undef;
 }
 
 sub error {
     my ($responder, $msg) = @_;
-    my $writer = $responder->([500, [ 'Content-Type' => 'text/plain',
-                                      'Content-Encoding' => 'UTF-8' ]]);
-    $writer->write($msg);
-    $writer->close;
+    my $writer = Geo::OGC::Service::XMLWriter::Caching->new('text/xml');
+    $writer->open_element('ExceptionReport', { version => "1.0" });
+    my $attributes = { exceptionCode => $msg->{exceptionCode} };
+    my $content;
+    $content = [ ExceptionText => $msg->{ExceptionText} ] if exists $msg->{ExceptionText};
+    if ($msg->{exceptionCode} eq 'MissingParameterValue') {
+        $attributes->{locator} = $msg->{locator};
+    }
+    $writer->element('Exception', $attributes, $content);
+    $writer->close_element;
+    $writer->stream($responder);
 }
 
 =pod
@@ -460,16 +501,35 @@ sub write {
     push @{$self->{cache}}, $line;
 }
 
+sub to_string {
+    my $self = shift;
+    my $xml = '<?xml version="1.0" encoding="UTF-8"?>';
+    for my $line (@{$self->{cache}}) {
+        $xml .= $line;
+    }
+    return $xml;
+}
+
 sub stream {
     my $self = shift;
     my $responder = shift;
+    my $debug = shift;
     my $writer = $responder->([200, [ 'Content-Type' => $self->{content_type},
                                       'Content-Encoding' => 'UTF-8' ]]);
     $writer->write('<?xml version="1.0" encoding="UTF-8"?>');
+    my $xml = '';
     for my $line (@{$self->{cache}}) {
         $writer->write($line);
+        $xml .= $line;
     }
     $writer->close;
+    if ($debug) {
+        my $parser = XML::LibXML->new(no_blanks => 1);
+        my $pp = XML::LibXML::PrettyPrint->new(indent_string => "  ");
+        my $dom = $parser->load_xml(string => $xml);
+        $pp->pretty_print($dom);
+        say STDERR $dom->toString;
+    }
 }
 
 1;
